@@ -1,7 +1,7 @@
 /*
    Various utilities
 
-   Copyright (C) 1994-2020
+   Copyright (C) 1994-2024
    Free Software Foundation, Inc.
 
    Written by:
@@ -35,12 +35,12 @@
 #include <config.h>
 
 #include <ctype.h>
+#include <stddef.h>             /* ptrdiff_t */
 #include <limits.h>
 #include <stdarg.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
-#include <sys/time.h>
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <unistd.h>
@@ -51,7 +51,6 @@
 #include "lib/vfs/vfs.h"
 #include "lib/strutil.h"
 #include "lib/util.h"
-#include "lib/timer.h"
 
 /*** global variables ****************************************************************************/
 
@@ -73,6 +72,8 @@
 
 /*** file scope type declarations ****************************************************************/
 
+/*** forward declarations (file scope functions) *************************************************/
+
 /*** file scope variables ************************************************************************/
 
 /* --------------------------------------------------------------------------------------------- */
@@ -110,7 +111,7 @@ is_8bit_printable (unsigned char c)
 /* --------------------------------------------------------------------------------------------- */
 
 static char *
-resolve_symlinks (const vfs_path_t * vpath)
+resolve_symlinks (const vfs_path_t *vpath)
 {
     char *p, *p2;
     char *buf, *buf2, *q, *r, c;
@@ -255,6 +256,9 @@ name_quote (const char *s, gboolean quote_percent)
 {
     GString *ret;
 
+    if (s == NULL || *s == '\0')
+        return NULL;
+
     ret = g_string_sized_new (64);
 
     if (*s == '-')
@@ -304,7 +308,7 @@ name_quote (const char *s, gboolean quote_percent)
         g_string_append_c (ret, *s);
     }
 
-    return g_string_free (ret, FALSE);
+    return g_string_free (ret, ret->len == 0);
 }
 
 /* --------------------------------------------------------------------------------------------- */
@@ -313,7 +317,8 @@ char *
 fake_name_quote (const char *s, gboolean quote_percent)
 {
     (void) quote_percent;
-    return g_strdup (s);
+
+    return (s == NULL || *s == '\0' ? NULL : g_strdup (s));
 }
 
 /* --------------------------------------------------------------------------------------------- */
@@ -327,15 +332,11 @@ const char *
 path_trunc (const char *path, size_t trunc_len)
 {
     vfs_path_t *vpath;
-    char *secure_path;
     const char *ret;
 
-    vpath = vfs_path_from_str (path);
-    secure_path = vfs_path_to_str_flags (vpath, 0, VPF_STRIP_PASSWORD);
-    vfs_path_free (vpath);
-
-    ret = str_trunc (secure_path, trunc_len);
-    g_free (secure_path);
+    vpath = vfs_path_from_str_flags (path, VPF_STRIP_PASSWORD);
+    ret = str_trunc (vfs_path_as_str (vpath), trunc_len);
+    vfs_path_free (vpath, TRUE);
 
     return ret;
 }
@@ -385,7 +386,7 @@ size_trunc_sep (uintmax_t size, gboolean use_si)
     d = x + sizeof (x) - 1;
     *d-- = '\0';
     /* @size format is "size unit", i.e. "[digits][space][letters]".
-       Copy all charactes after digits. */
+       Copy all characters after digits. */
     while (p >= y && !g_ascii_isdigit (*p))
         *d-- = *p--;
     for (count = 0; p >= y; count++)
@@ -451,9 +452,12 @@ size_trunc_len (char *buffer, unsigned int len, uintmax_t size, int units, gbool
      */
 #endif
     };
+
+    static const char *const suffix[] =
+        { "", "K", "M", "G", "T", "P", "E", "Z", "Y", "R", "Q", NULL };
+    static const char *const suffix_lc[] =
+        { "", "k", "m", "g", "t", "p", "e", "z", "y", "r", "q", NULL };
     /* *INDENT-ON* */
-    static const char *const suffix[] = { "", "K", "M", "G", "T", "P", "E", "Z", "Y", NULL };
-    static const char *const suffix_lc[] = { "", "k", "m", "g", "t", "p", "e", "z", "y", NULL };
 
     const char *const *sfx = use_si ? suffix_lc : suffix;
     int j = 0;
@@ -590,7 +594,7 @@ extension (const char *filename)
 
 char *
 load_mc_home_file (const char *from, const char *filename, char **allocated_filename,
-                   size_t * length)
+                   size_t *length)
 {
     char *hintfile_base, *hintfile;
     char *lang;
@@ -664,7 +668,7 @@ x_basename (const char *s)
     {
         /* avoid trailing PATH_SEP, if present */
         if (!IS_PATH_SEP (s[strlen (s) - 1]))
-            return (path_sep != NULL) ? path_sep + 1 : s;
+            return path_sep + 1;
 
         while (--path_sep > s && !IS_PATH_SEP (*path_sep))
             ;
@@ -783,6 +787,7 @@ strip_ctrl_codes (char *s)
                             r = new_r + 1;
                             goto osc_out;
                         }
+                        break;
                     default:
                         break;
                     }
@@ -829,27 +834,25 @@ get_compression_type (int fd, const char *name)
         return COMPRESSION_NONE;
 
     /* GZIP_MAGIC and OLD_GZIP_MAGIC */
-    if (magic[0] == 037 && (magic[1] == 0213 || magic[1] == 0236))
+    if (magic[0] == 0x1F && (magic[1] == 0x8B || magic[1] == 0x9E))
         return COMPRESSION_GZIP;
 
     /* PKZIP_MAGIC */
-    if (magic[0] == 0120 && magic[1] == 0113 && magic[2] == 003 && magic[3] == 004)
+    if (magic[0] == 'P' && magic[1] == 'K' && magic[2] == 0x03 && magic[3] == 0x04)
     {
         /* Read compression type */
         mc_lseek (fd, 8, SEEK_SET);
         if (mc_read (fd, (char *) magic, 2) != 2)
             return COMPRESSION_NONE;
 
-        /* Gzip can handle only deflated (8) or stored (0) files */
         if ((magic[0] != 8 && magic[0] != 0) || magic[1] != 0)
             return COMPRESSION_NONE;
 
-        /* Compatible with gzip */
-        return COMPRESSION_GZIP;
+        return COMPRESSION_ZIP;
     }
 
     /* PACK_MAGIC and LZH_MAGIC and compress magic */
-    if (magic[0] == 037 && (magic[1] == 036 || magic[1] == 0240 || magic[1] == 0235))
+    if (magic[0] == 0x1F && (magic[1] == 0x1E || magic[1] == 0xA0 || magic[1] == 0x9D))
         /* Compatible with gzip */
         return COMPRESSION_GZIP;
 
@@ -885,6 +888,11 @@ get_compression_type (int fd, const char *name)
         && magic[2] == 'Z' && magic[3] == 'M' && magic[4] == 'A' && magic[5] == 0x00)
         return COMPRESSION_LZMA;
 
+    /* LZO format - \x89\x4c\x5a\x4f\x00\x0d\x0a\x1a\x0a    lzop compressed data */
+    if (magic[0] == 0x89 && magic[1] == 0x4c &&
+        magic[2] == 0x5a && magic[3] == 0x4f && magic[4] == 0x00 && magic[5] == 0x0d)
+        return COMPRESSION_LZO;
+
     /* XZ compression magic */
     if (magic[0] == 0xFD
         && magic[1] == 0x37
@@ -895,7 +903,7 @@ get_compression_type (int fd, const char *name)
         return COMPRESSION_ZSTD;
 
     str_len = strlen (name);
-    /* HACK: we must belive to extension of LZMA file :) ... */
+    /* HACK: we must believe to extension of LZMA file :) ... */
     if ((str_len > 5 && strcmp (&name[str_len - 5], ".lzma") == 0) ||
         (str_len > 4 && strcmp (&name[str_len - 4], ".tlz") == 0))
         return COMPRESSION_LZMA;
@@ -910,6 +918,8 @@ decompress_extension (int type)
 {
     switch (type)
     {
+    case COMPRESSION_ZIP:
+        return "/uz" VFS_PATH_URL_DELIMITER;
     case COMPRESSION_GZIP:
         return "/ugz" VFS_PATH_URL_DELIMITER;
     case COMPRESSION_BZIP:
@@ -922,6 +932,8 @@ decompress_extension (int type)
         return "/ulz4" VFS_PATH_URL_DELIMITER;
     case COMPRESSION_LZMA:
         return "/ulzma" VFS_PATH_URL_DELIMITER;
+    case COMPRESSION_LZO:
+        return "/ulzo" VFS_PATH_URL_DELIMITER;
     case COMPRESSION_XZ:
         return "/uxz" VFS_PATH_URL_DELIMITER;
     case COMPRESSION_ZSTD:
@@ -1014,7 +1026,7 @@ convert_controls (const char *p)
  */
 
 char *
-diff_two_paths (const vfs_path_t * vpath1, const vfs_path_t * vpath2)
+diff_two_paths (const vfs_path_t *vpath1, const vfs_path_t *vpath2)
 {
     int j, prevlen = -1, currlen;
     char *my_first = NULL, *my_second = NULL;
@@ -1085,7 +1097,7 @@ diff_two_paths (const vfs_path_t * vpath1, const vfs_path_t * vpath2)
  */
 
 GList *
-list_append_unique (GList * list, char *text)
+list_append_unique (GList *list, char *text)
 {
     GList *lc_link;
 
@@ -1125,8 +1137,8 @@ list_append_unique (GList * list, char *text)
  */
 
 void
-load_file_position (const vfs_path_t * filename_vpath, long *line, long *column, off_t * offset,
-                    GArray ** bookmarks)
+load_file_position (const vfs_path_t *filename_vpath, long *line, long *column, off_t *offset,
+                    GArray **bookmarks)
 {
     char *fn;
     FILE *f;
@@ -1216,8 +1228,8 @@ load_file_position (const vfs_path_t * filename_vpath, long *line, long *column,
  */
 
 void
-save_file_position (const vfs_path_t * filename_vpath, long line, long column, off_t offset,
-                    GArray * bookmarks)
+save_file_position (const vfs_path_t *filename_vpath, long line, long column, off_t offset,
+                    GArray *bookmarks)
 {
     static size_t filepos_max_saved_entries = 0;
     char *fn, *tmp_fn;
@@ -1260,7 +1272,7 @@ save_file_position (const vfs_path_t * filename_vpath, long line, long column, o
         if (bookmarks != NULL)
             for (i = 0; i < bookmarks->len && i < MAX_SAVED_BOOKMARKS; i++)
                 if (fprintf (f, ";%zu", g_array_index (bookmarks, size_t, i)) < 0)
-                    goto write_position_error;
+                      goto write_position_error;
 
         if (fprintf (f, "\n") < 0)
             goto write_position_error;
@@ -1384,7 +1396,7 @@ mc_util_unlink_backup_if_possible (const char *file_name, const char *backup_suf
 
         vpath = vfs_path_from_str (backup_path);
         mc_unlink (vpath);
-        vfs_path_free (vpath);
+        vfs_path_free (vpath, TRUE);
     }
 
     g_free (backup_path);
@@ -1463,7 +1475,7 @@ mc_get_profile_root (void)
  */
 
 void
-mc_propagate_error (GError ** dest, int code, const char *format, ...)
+mc_propagate_error (GError **dest, int code, const char *format, ...)
 {
     if (dest != NULL && *dest == NULL)
     {
@@ -1489,7 +1501,7 @@ mc_propagate_error (GError ** dest, int code, const char *format, ...)
  */
 
 void
-mc_replace_error (GError ** dest, int code, const char *format, ...)
+mc_replace_error (GError **dest, int code, const char *format, ...)
 {
     if (dest != NULL)
     {
@@ -1513,16 +1525,16 @@ mc_replace_error (GError ** dest, int code, const char *format, ...)
  * and if it has then updates the timestamp.
  *
  * @param timestamp the last timestamp in microseconds, updated if the given time elapsed
- * @param deleay amount of time in microseconds
+ * @param delay amount of time in microseconds
 
  * @return TRUE if clock skew detected, FALSE otherwise
  */
 gboolean
-mc_time_elapsed (guint64 * timestamp, guint64 delay)
+mc_time_elapsed (gint64 *timestamp, gint64 delay)
 {
-    guint64 now;
+    gint64 now;
 
-    now = mc_timer_elapsed (mc_global.timer);
+    now = g_get_monotonic_time ();
 
     if (now >= *timestamp && now < *timestamp + delay)
         return FALSE;
