@@ -1,7 +1,7 @@
 /*
    Various utilities - Unix variants
 
-   Copyright (C) 1994-2020
+   Copyright (C) 1994-2024
    Free Software Foundation, Inc.
 
    Written by:
@@ -10,6 +10,7 @@
    Dugan Porter, 1994, 1995, 1996
    Jakub Jelinek, 1994, 1995, 1996
    Mauricio Plaza, 1994, 1995, 1996
+   Andrew Borodin <aborodin@vmail.ru> 2010-2024
 
    The mc_realpath routine is mostly from uClibc package, written
    by Rick Sladkey <jrs@world.std.com>
@@ -53,12 +54,6 @@
 #include <sys/select.h>
 #endif
 #include <sys/wait.h>
-#ifdef HAVE_SYS_IOCTL_H
-#include <sys/ioctl.h>
-#endif
-#ifdef HAVE_GET_PROCESS_STATS
-#include <sys/procstats.h>
-#endif
 #include <pwd.h>
 #include <grp.h>
 
@@ -66,7 +61,7 @@
 
 #include "lib/unixcompat.h"
 #include "lib/vfs/vfs.h"        /* VFS_ENCODING_PREFIX */
-#include "lib/strutil.h"        /* str_move() */
+#include "lib/strutil.h"        /* str_move(), str_tokenize() */
 #include "lib/util.h"
 #include "lib/widget.h"         /* message() */
 #include "lib/vfs/xdirentry.h"
@@ -75,8 +70,6 @@
 #include "lib/charsets.h"
 #endif
 
-#include "utilunix.h"
-
 /*** global variables ****************************************************************************/
 
 struct sigaction startup_handler;
@@ -86,10 +79,6 @@ struct sigaction startup_handler;
 #define UID_CACHE_SIZE 200
 #define GID_CACHE_SIZE 30
 
-/* Pipes are guaranteed to be able to hold at least 4096 bytes */
-/* More than that would be unportable */
-#define MAX_PIPE_SIZE 4096
-
 /*** file scope type declarations ****************************************************************/
 
 typedef struct
@@ -112,20 +101,19 @@ typedef struct
     struct sigaction stop;
 } my_system_sigactions_t;
 
+/*** forward declarations (file scope functions) *************************************************/
+
 /*** file scope variables ************************************************************************/
 
 static int_cache uid_cache[UID_CACHE_SIZE];
 static int_cache gid_cache[GID_CACHE_SIZE];
 
-static int error_pipe[2];       /* File descriptors of error pipe */
-static int old_error;           /* File descriptor of old standard error */
-
 /* --------------------------------------------------------------------------------------------- */
 /*** file scope functions ************************************************************************/
 /* --------------------------------------------------------------------------------------------- */
 
 static char *
-i_cache_match (int id, int_cache * cache, int size)
+i_cache_match (int id, int_cache *cache, int size)
 {
     int i;
 
@@ -138,7 +126,7 @@ i_cache_match (int id, int_cache * cache, int size)
 /* --------------------------------------------------------------------------------------------- */
 
 static void
-i_cache_add (int id, int_cache * cache, int size, char *text, int *last)
+i_cache_add (int id, int_cache *cache, int size, char *text, int *last)
 {
     g_free (cache[*last].string);
     cache[*last].string = g_strdup (text);
@@ -179,7 +167,7 @@ my_fork (void)
 /* --------------------------------------------------------------------------------------------- */
 
 static void
-my_system__save_sigaction_handlers (my_system_sigactions_t * sigactions)
+my_system__save_sigaction_handlers (my_system_sigactions_t *sigactions)
 {
     struct sigaction ignore;
 
@@ -198,7 +186,7 @@ my_system__save_sigaction_handlers (my_system_sigactions_t * sigactions)
 /* --------------------------------------------------------------------------------------------- */
 
 static void
-my_system__restore_sigaction_handlers (my_system_sigactions_t * sigactions)
+my_system__restore_sigaction_handlers (my_system_sigactions_t *sigactions)
 {
     sigaction (SIGINT, &sigactions->intr, NULL);
     sigaction (SIGQUIT, &sigactions->quit, NULL);
@@ -208,37 +196,31 @@ my_system__restore_sigaction_handlers (my_system_sigactions_t * sigactions)
 /* --------------------------------------------------------------------------------------------- */
 
 static GPtrArray *
-my_system_make_arg_array (int flags, const char *shell, char **execute_name)
+my_system_make_arg_array (int flags, const char *shell)
 {
     GPtrArray *args_array;
 
-    args_array = g_ptr_array_new ();
-
     if ((flags & EXECUTE_AS_SHELL) != 0)
     {
+        args_array = g_ptr_array_new ();
         g_ptr_array_add (args_array, (gpointer) shell);
         g_ptr_array_add (args_array, (gpointer) "-c");
-        *execute_name = g_strdup (shell);
+    }
+    else if (shell == NULL || *shell == '\0')
+    {
+        args_array = g_ptr_array_new ();
+        g_ptr_array_add (args_array, NULL);
     }
     else
-    {
-        char *shell_token;
+        args_array = str_tokenize (shell);
 
-        shell_token = shell != NULL ? strchr (shell, ' ') : NULL;
-        if (shell_token == NULL)
-            *execute_name = g_strdup (shell);
-        else
-            *execute_name = g_strndup (shell, (gsize) (shell_token - shell));
-
-        g_ptr_array_add (args_array, (gpointer) shell);
-    }
     return args_array;
 }
 
 /* --------------------------------------------------------------------------------------------- */
 
 static void
-mc_pread_stream (mc_pipe_stream_t * ps, const fd_set * fds)
+mc_pread_stream (mc_pipe_stream_t *ps, const fd_set *fds)
 {
     size_t buf_len;
     ssize_t read_len;
@@ -277,6 +259,8 @@ mc_pread_stream (mc_pipe_stream_t * ps, const fd_set * fds)
         if (ps->null_term)
             ps->buf[(size_t) ps->len] = '\0';
     }
+
+    ps->pos = 0;
 }
 
 /* --------------------------------------------------------------------------------------------- */
@@ -373,7 +357,7 @@ my_exit (int status)
  *                    Shell (or command) will be found in paths described in PATH variable
  *                    (if shell parameter doesn't begin from path delimiter)
  * @parameter command Command for shell (or first parameter for command, if flags contain EXECUTE_AS_SHELL)
- * @return 0 if successfull, -1 otherwise
+ * @return 0 if successful, -1 otherwise
  */
 
 int
@@ -388,12 +372,12 @@ my_system (int flags, const char *shell, const char *command)
  *
  * @parameter flags addition conditions for running external programs.
  * @parameter shell shell (if flags contain EXECUTE_AS_SHELL), command to run otherwise.
- *                  Shell (or command) will be found in pathes described in PATH variable
+ *                  Shell (or command) will be found in paths described in PATH variable
  *                  (if shell parameter doesn't begin from path delimiter)
  * @parameter ...   Command for shell with addition parameters for shell
  *                  (or parameters for command, if flags contain EXECUTE_AS_SHELL).
  *                  Should be NULL terminated.
- * @return 0 if successfull, -1 otherwise
+ * @return 0 if successful, -1 otherwise
  */
 
 int
@@ -426,7 +410,7 @@ my_systeml (int flags, const char *shell, ...)
  * @parameter command command to run. Command will be found in paths described in PATH variable
  *                    (if command parameter doesn't begin from path delimiter)
  * @parameter argv    Array of strings (NULL-terminated) with parameters for command
- * @return 0 if successfull, -1 otherwise
+ * @return 0 if successful, -1 otherwise
  */
 
 int
@@ -474,17 +458,19 @@ my_systemv (const char *command, char *const argv[])
  *                    Shell (or command) will be found in paths described in PATH variable
  *                    (if shell parameter doesn't begin from path delimiter)
  * @parameter argv    Array of strings (NULL-terminated) with parameters for command
- * @return 0 if successfull, -1 otherwise
+ * @return 0 if successful, -1 otherwise
  */
 
 int
 my_systemv_flags (int flags, const char *command, char *const argv[])
 {
-    char *execute_name = NULL;
+    const char *execute_name;
     GPtrArray *args_array;
     int status = 0;
 
-    args_array = my_system_make_arg_array (flags, command, &execute_name);
+    args_array = my_system_make_arg_array (flags, command);
+
+    execute_name = g_ptr_array_index (args_array, 0);
 
     for (; argv != NULL && *argv != NULL; argv++)
         g_ptr_array_add (args_array, *argv);
@@ -492,7 +478,6 @@ my_systemv_flags (int flags, const char *command, char *const argv[])
     g_ptr_array_add (args_array, NULL);
     status = my_systemv (execute_name, (char *const *) args_array->pdata);
 
-    g_free (execute_name);
     g_ptr_array_free (args_array, TRUE);
 
     return status;
@@ -503,13 +488,15 @@ my_systemv_flags (int flags, const char *command, char *const argv[])
  * Create pipe and run child process.
  *
  * @parameter command command line of child process
- * @paremeter error contains pointer to object to handle error code and message
+ * @parameter read_out do or don't read the stdout of child process
+ * @parameter read_err do or don't read the stderr of child process
+ * @parameter error contains pointer to object to handle error code and message
  *
  * @return newly created object of mc_pipe_t class in success, NULL otherwise
  */
 
 mc_pipe_t *
-mc_popen (const char *command, GError ** error)
+mc_popen (const char *command, gboolean read_out, gboolean read_err, GError **error)
 {
     mc_pipe_t *p;
     const char *const argv[] = { "/bin/sh", "sh", "-c", command, NULL };
@@ -522,9 +509,13 @@ mc_popen (const char *command, GError ** error)
         goto ret_err;
     }
 
+    p->out.fd = -1;
+    p->err.fd = -1;
+
     if (!g_spawn_async_with_pipes
-        (NULL, (gchar **) argv, NULL, G_SPAWN_DO_NOT_REAP_CHILD | G_SPAWN_FILE_AND_ARGV_ZERO,
-         NULL, NULL, &p->child_pid, NULL, &p->out.fd, &p->err.fd, error))
+        (NULL, (gchar **) argv, NULL, G_SPAWN_DO_NOT_REAP_CHILD | G_SPAWN_FILE_AND_ARGV_ZERO, NULL,
+         NULL, &p->child_pid, NULL, read_out ? &p->out.fd : NULL, read_err ? &p->err.fd : NULL,
+         error))
     {
         mc_replace_error (error, MC_PIPE_ERROR_CREATE_PIPE_STREAM, "%s",
                           _("Cannot create pipe streams"));
@@ -553,9 +544,9 @@ mc_popen (const char *command, GError ** error)
  * @parameter p pipe descriptor
  *
  * The lengths of read data contain in p->out.len and p->err.len.
- * Before read, p->xxx.len is an input:
- *   p->xxx.len > 0:  do read stream p->xxx and store data in p->xxx.buf;
- *   p->xxx.len <= 0: do not read stream p->xxx.
+ *
+ * Before read, p->xxx.len is an input. It defines the number of data to read.
+ * Should not be greater than MC_PIPE_BUFSIZE.
  *
  * After read, p->xxx.len is an output and contains the following:
  *   p->xxx.len > 0: an actual length of read data stored in p->xxx.buf;
@@ -563,11 +554,11 @@ mc_popen (const char *command, GError ** error)
  *   p->xxx.len == MC_PIPE_STREAM_UNREAD: stream p->xxx was not read;
  *   p->xxx.len == MC_PIPE_ERROR_READ: reading error, and p->xxx.errno is set appropriately.
  *
- * @paremeter error contains pointer to object to handle error code and message
+ * @parameter error contains pointer to object to handle error code and message
  */
 
 void
-mc_pread (mc_pipe_t * p, GError ** error)
+mc_pread (mc_pipe_t *p, GError **error)
 {
     gboolean read_out, read_err;
     fd_set fds;
@@ -577,8 +568,8 @@ mc_pread (mc_pipe_t * p, GError ** error)
     if (error != NULL)
         *error = NULL;
 
-    read_out = p->out.fd >= 0 && p->out.len > 0;
-    read_err = p->err.fd >= 0 && p->err.len > 0;
+    read_out = p->out.fd >= 0;
+    read_err = p->err.fd >= 0;
 
     if (!read_out && !read_err)
     {
@@ -622,19 +613,70 @@ mc_pread (mc_pipe_t * p, GError ** error)
         p->err.len = MC_PIPE_STREAM_UNREAD;
 }
 
+/* --------------------------------------------------------------------------------------------- */
+/**
+ * Reads a line from @stream. Reading stops after an EOL or a newline. If a newline is read,
+ * it is appended to the line.
+ *
+ * @stream mc_pipe_stream_t object
+ *
+ * @return newly created GString or NULL in case of EOL;
+ */
+
+GString *
+mc_pstream_get_string (mc_pipe_stream_t *ps)
+{
+    char *s;
+    size_t size, i;
+    gboolean escape = FALSE;
+
+    g_return_val_if_fail (ps != NULL, NULL);
+
+    if (ps->len < 0)
+        return NULL;
+
+    size = ps->len - ps->pos;
+
+    if (size == 0)
+        return NULL;
+
+    s = ps->buf + ps->pos;
+
+    if (s[0] == '\0')
+        return NULL;
+
+    /* find '\0' or unescaped '\n' */
+    for (i = 0; i < size && !(s[i] == '\0' || (s[i] == '\n' && !escape)); i++)
+        escape = s[i] == '\\' ? !escape : FALSE;
+
+    if (i != size && s[i] == '\n')
+        i++;
+
+    ps->pos += i;
+
+    return g_string_new_len (s, i);
+}
+
 /* --------------------------------------------------------------------------------------------- */
 /**
  * Close pipe and destroy pipe descriptor.
  *
- * @paremeter p pipe descriptor
- * @paremeter error contains pointer to object to handle error code and message
+ * @parameter p pipe descriptor
+ * @parameter error contains pointer to object to handle error code and message
  */
 
 void
-mc_pclose (mc_pipe_t * p, GError ** error)
+mc_pclose (mc_pipe_t *p, GError **error)
 {
     int res;
 
+    if (p == NULL)
+    {
+        mc_replace_error (error, MC_PIPE_ERROR_READ, "%s",
+                          _("Cannot close pipe descriptor (p == NULL)"));
+        return;
+    }
+
     if (p->out.fd >= 0)
         res = close (p->out.fd);
     if (p->err.fd >= 0)
@@ -707,128 +749,17 @@ tilde_expand (const char *directory)
 
 /* --------------------------------------------------------------------------------------------- */
 /**
- * Creates a pipe to hold standard error for a later analysis.
- * The pipe can hold 4096 bytes. Make sure no more is written
- * or a deadlock might occur.
- */
-
-void
-open_error_pipe (void)
-{
-    if (pipe (error_pipe) < 0)
-        message (D_NORMAL, _("Warning"), _("Pipe failed"));
-
-    old_error = dup (STDERR_FILENO);
-    if (old_error < 0 || close (STDERR_FILENO) != 0 || dup (error_pipe[1]) != STDERR_FILENO)
-    {
-        message (D_NORMAL, _("Warning"), _("Dup failed"));
-
-        close (error_pipe[0]);
-        error_pipe[0] = -1;
-    }
-    else
-    {
-        /*
-         * Settng stderr in nonblocking mode as we close it earlier, than
-         * program stops. We try to read some error at program startup,
-         * but we should not block on it.
-         *
-         * TODO: make piped stdin/stderr poll()/select()able to get rid
-         * of following hack.
-         */
-        int fd_flags;
-
-        fd_flags = fcntl (error_pipe[0], F_GETFL, NULL);
-        if (fd_flags != -1)
-        {
-            fd_flags |= O_NONBLOCK;
-
-            if (fcntl (error_pipe[0], F_SETFL, fd_flags) == -1)
-            {
-                /* TODO: handle it somehow */
-            }
-        }
-    }
-    /* we never write there */
-    close (error_pipe[1]);
-    error_pipe[1] = -1;
-}
-
-/* --------------------------------------------------------------------------------------------- */
-/**
- * Close a pipe
+ * Canonicalize path.
  *
- * @param error '-1' - ignore errors, '0' - display warning, '1' - display error
- * @param text is prepended to the error message from the pipe
+ * @param path path to file
+ * @param flags canonicalization flags
  *
- * @return not 0 if an error was displayed
- */
-
-int
-close_error_pipe (int error, const char *text)
-{
-    const char *title;
-    char msg[MAX_PIPE_SIZE];
-    int len = 0;
-
-    /* already closed */
-    if (error_pipe[0] == -1)
-        return 0;
-
-    if (error < 0 || (error > 0 && (error & D_ERROR) != 0))
-        title = MSG_ERROR;
-    else
-        title = _("Warning");
-    if (old_error >= 0)
-    {
-        if (dup2 (old_error, STDERR_FILENO) == -1)
-        {
-            if (error < 0)
-                error = D_ERROR;
-
-            message (error, MSG_ERROR, _("Error dup'ing old error pipe"));
-            return 1;
-        }
-        close (old_error);
-        len = read (error_pipe[0], msg, sizeof (msg) - 1);
-
-        if (len >= 0)
-            msg[len] = 0;
-        close (error_pipe[0]);
-        error_pipe[0] = -1;
-    }
-    if (error < 0)
-        return 0;               /* Just ignore error message */
-    if (text == NULL)
-    {
-        if (len <= 0)
-            return 0;           /* Nothing to show */
-
-        /* Show message from pipe */
-        message (error, title, "%s", msg);
-    }
-    else
-    {
-        /* Show given text and possible message from pipe */
-        message (error, title, "%s\n%s", text, msg);
-    }
-    return 1;
-}
-
-/* --------------------------------------------------------------------------------------------- */
-/**
- * Canonicalize path, and return a new path.  Do everything in place.
- * The new path differs from path in:
- *      Multiple '/'s are collapsed to a single '/'.
- *      Leading './'s and trailing '/.'s are removed.
- *      Trailing '/'s are removed.
- *      Non-leading '../'s and trailing '..'s are handled by removing
- *      portions of the path.
+ * All modifications of @path are made in place.
  * Well formed UNC paths are modified only in the local part.
  */
 
 void
-custom_canonicalize_pathname (char *path, CANON_PATH_FLAGS flags)
+canonicalize_pathname_custom (char *path, canon_path_flags_t flags)
 {
     char *p, *s;
     char *lpath = path;         /* path without leading UNC part */
@@ -1050,26 +981,6 @@ custom_canonicalize_pathname (char *path, CANON_PATH_FLAGS flags)
 
 /* --------------------------------------------------------------------------------------------- */
 
-void
-canonicalize_pathname (char *path)
-{
-    custom_canonicalize_pathname (path, CANON_PATH_ALL);
-}
-
-/* --------------------------------------------------------------------------------------------- */
-
-#ifdef HAVE_GET_PROCESS_STATS
-int
-gettimeofday (struct timeval *tp, void *tzp)
-{
-    (void) tzp;
-
-    return get_process_stats (tp, PS_SELF, 0, 0);
-}
-#endif /* HAVE_GET_PROCESS_STATS */
-
-/* --------------------------------------------------------------------------------------------- */
-
 char *
 mc_realpath (const char *path, char *resolved_path)
 {
@@ -1308,13 +1219,13 @@ mc_build_filenamev (const char *first_element, va_list args)
     GString *path;
     char *ret;
 
-    if (element == NULL)
+    if (first_element == NULL)
         return NULL;
 
-    path = g_string_new ("");
-
     absolute = IS_PATH_SEP (*first_element);
 
+    path = g_string_new (absolute ? PATH_SEP_STR : "");
+
     do
     {
         if (*element == '\0')
@@ -1322,7 +1233,6 @@ mc_build_filenamev (const char *first_element, va_list args)
         else
         {
             char *tmp_element;
-            size_t len;
             const char *start;
 
             tmp_element = g_strdup (element);
@@ -1330,11 +1240,10 @@ mc_build_filenamev (const char *first_element, va_list args)
             element = va_arg (args, char *);
 
             canonicalize_pathname (tmp_element);
-            len = strlen (tmp_element);
             start = IS_PATH_SEP (tmp_element[0]) ? tmp_element + 1 : tmp_element;
 
             g_string_append (path, start);
-            if (!IS_PATH_SEP (tmp_element[len - 1]) && element != NULL)
+            if (!IS_PATH_SEP (path->str[path->len - 1]) && element != NULL)
                 g_string_append_c (path, PATH_SEP);
 
             g_free (tmp_element);
@@ -1342,9 +1251,6 @@ mc_build_filenamev (const char *first_element, va_list args)
     }
     while (element != NULL);
 
-    if (absolute)
-        g_string_prepend_c (path, PATH_SEP);
-
     ret = g_string_free (path, FALSE);
     canonicalize_pathname (ret);
 
