@@ -1,7 +1,7 @@
 /*
    Chown-advanced command -- for the Midnight Commander
 
-   Copyright (C) 1994-2020
+   Copyright (C) 1994-2024
    Free Software Foundation, Inc.
 
    This file is part of the Midnight Commander.
@@ -45,9 +45,7 @@
 #include "lib/util.h"
 #include "lib/widget.h"
 
-#include "midnight.h"           /* current_panel */
-
-#include "achown.h"
+#include "cmd.h"                /* advanced_chown_cmd() */
 
 /*** global variables ****************************************************************************/
 
@@ -64,6 +62,8 @@
 
 /*** file scope type declarations ****************************************************************/
 
+/*** forward declarations (file scope functions) *************************************************/
+
 /*** file scope variables ************************************************************************/
 
 static struct
@@ -74,8 +74,7 @@ static struct
     int x;
     int len;
     const char *text;
-} advanced_chown_but[BUTTONS] =
-{
+} advanced_chown_but[BUTTONS] = {
     /* *INDENT-OFF* */
     { 0, B_ENTER,   NARROW_BUTTON,  3, 0, "   " },
     { 0, B_ENTER,   NARROW_BUTTON, 11, 0, "   " },
@@ -238,7 +237,7 @@ update_ownership (void)
 /* --------------------------------------------------------------------------------------------- */
 
 static void
-print_flags (const WDialog * h)
+print_flags (const WDialog *h)
 {
     int i;
 
@@ -279,7 +278,7 @@ print_flags (const WDialog * h)
 /* --------------------------------------------------------------------------------------------- */
 
 static void
-advanced_chown_refresh (WDialog * h)
+advanced_chown_refresh (const WDialog *h)
 {
     tty_setcolor (COLOR_NORMAL);
 
@@ -315,7 +314,7 @@ advanced_chown_info_update (void)
 /* --------------------------------------------------------------------------------------------- */
 
 static void
-update_mode (WGroup * g)
+update_mode (WGroup *g)
 {
     print_flags (DIALOG (g));
     advanced_chown_info_update ();
@@ -325,7 +324,7 @@ update_mode (WGroup * g)
 /* --------------------------------------------------------------------------------------------- */
 
 static cb_ret_t
-perm_button_callback (Widget * w, Widget * sender, widget_msg_t msg, int parm, void *data)
+perm_button_callback (Widget *w, Widget *sender, widget_msg_t msg, int parm, void *data)
 {
     WButton *b = BUTTON (w);
     WGroup *g = w->owner;
@@ -439,7 +438,7 @@ perm_button_callback (Widget * w, Widget * sender, widget_msg_t msg, int parm, v
 /* --------------------------------------------------------------------------------------------- */
 
 static void
-perm_button_mouse_callback (Widget * w, mouse_msg_t msg, mouse_event_t * event)
+perm_button_mouse_callback (Widget *w, mouse_msg_t msg, mouse_event_t *event)
 {
     switch (msg)
     {
@@ -479,7 +478,7 @@ perm_button_new (int y, int x, int action, button_flags_t flags, const char *tex
 /* --------------------------------------------------------------------------------------------- */
 
 static cb_ret_t
-chl_callback (Widget * w, Widget * sender, widget_msg_t msg, int parm, void *data)
+chl_callback (Widget *w, Widget *sender, widget_msg_t msg, int parm, void *data)
 {
     switch (msg)
     {
@@ -492,8 +491,9 @@ chl_callback (Widget * w, Widget * sender, widget_msg_t msg, int parm, void *dat
                 WDialog *h = DIALOG (w);
 
                 h->ret_value = parm;
-                dlg_stop (h);
+                dlg_close (h);
             }
+            break;
         default:
             break;
         }
@@ -507,7 +507,7 @@ chl_callback (Widget * w, Widget * sender, widget_msg_t msg, int parm, void *dat
 /* --------------------------------------------------------------------------------------------- */
 
 static int
-user_group_button_cb (WButton * button, int action)
+user_group_button_cb (WButton *button, int action)
 {
     Widget *w = WIDGET (button);
     int f_pos;
@@ -530,7 +530,7 @@ user_group_button_cb (WButton * button, int action)
 
         gboolean is_owner = (f_pos == BUTTONS_PERM - 2);
         const char *title;
-        int lxx, b_pos;
+        int lxx, b_current;
         WDialog *chl_dlg;
         WListbox *chl_list;
         int result;
@@ -543,22 +543,22 @@ user_group_button_cb (WButton * button, int action)
         if (is_owner)
         {
             title = _("owner");
-            lxx = WIDGET (b_user)->x + 1;
+            lxx = WIDGET (b_user)->rect.x + 1;
         }
         else
         {
             title = _("group");
-            lxx = WIDGET (b_group)->x + 1;
+            lxx = WIDGET (b_group)->rect.x + 1;
         }
 
         chl_dlg =
-            dlg_create (TRUE, wh->y - 1, lxx, wh->lines + 2, 17, WPOS_KEEP_DEFAULT, TRUE,
+            dlg_create (TRUE, wh->rect.y - 1, lxx, wh->rect.lines + 2, 17, WPOS_KEEP_DEFAULT, TRUE,
                         dialog_colors, chl_callback, NULL, "[Advanced Chown]", title);
 
         /* get new listboxes */
         chl_list =
-            listbox_new (1, 1, WIDGET (chl_dlg)->lines - 2, WIDGET (chl_dlg)->cols - 2, FALSE,
-                         NULL);
+            listbox_new (1, 1, WIDGET (chl_dlg)->rect.lines - 2, WIDGET (chl_dlg)->rect.cols - 2,
+                         FALSE, NULL);
         listbox_add_item (chl_list, LISTBOX_APPEND_AT_END, 0, "<Unknown>", NULL, FALSE);
         if (is_owner)
         {
@@ -581,16 +581,16 @@ user_group_button_cb (WButton * button, int action)
             fe = listbox_search_text (chl_list, get_group (sf_stat.st_gid));
         }
 
-        listbox_select_entry (chl_list, fe);
+        listbox_set_current (chl_list, fe);
 
-        b_pos = chl_list->pos;
+        b_current = chl_list->current;
         group_add_widget (GROUP (chl_dlg), chl_list);
 
         result = dlg_run (chl_dlg);
 
         if (result != B_CANCEL)
         {
-            if (b_pos != chl_list->pos)
+            if (b_current != chl_list->current)
             {
                 gboolean ok = FALSE;
                 char *text;
@@ -643,7 +643,7 @@ user_group_button_cb (WButton * button, int action)
         }
 
         /* Here we used to redraw the window */
-        dlg_destroy (chl_dlg);
+        widget_destroy (WIDGET (chl_dlg));
     }
     while (chl_end);
 
@@ -653,7 +653,7 @@ user_group_button_cb (WButton * button, int action)
 /* --------------------------------------------------------------------------------------------- */
 
 static cb_ret_t
-advanced_chown_bg_callback (Widget * w, Widget * sender, widget_msg_t msg, int parm, void *data)
+advanced_chown_bg_callback (Widget *w, Widget *sender, widget_msg_t msg, int parm, void *data)
 {
     switch (msg)
     {
@@ -671,7 +671,7 @@ advanced_chown_bg_callback (Widget * w, Widget * sender, widget_msg_t msg, int p
 /* --------------------------------------------------------------------------------------------- */
 
 static cb_ret_t
-advanced_chown_callback (Widget * w, Widget * sender, widget_msg_t msg, int parm, void *data)
+advanced_chown_callback (Widget *w, Widget *sender, widget_msg_t msg, int parm, void *data)
 {
     WGroup *g = GROUP (w);
     int i = 0;
@@ -728,7 +728,7 @@ advanced_chown_callback (Widget * w, Widget * sender, widget_msg_t msg, int parm
 /* --------------------------------------------------------------------------------------------- */
 
 static WDialog *
-advanced_chown_dlg_create (void)
+advanced_chown_dlg_create (WPanel *panel)
 {
     gboolean single_set;
     WDialog *ch_dlg;
@@ -742,7 +742,7 @@ advanced_chown_dlg_create (void)
     flag_pos = 0;
     x_toggle = 070;
 
-    single_set = (current_panel->marked < 2);
+    single_set = (panel->marked < 2);
     if (!single_set)
         lines += 2;
 
@@ -754,7 +754,7 @@ advanced_chown_dlg_create (void)
     /* draw background */
     ch_dlg->bg->callback = advanced_chown_bg_callback;
 
-    l_filename = label_new (2, 3, "");
+    l_filename = label_new (2, 3, NULL);
     group_add_widget (ch_grp, l_filename);
 
     group_add_widget (ch_grp, hline_new (3, -1, -1));
@@ -773,7 +773,7 @@ advanced_chown_dlg_create (void)
     b_group = button_new (XTRACT (4, BY, user_group_button_cb));
     advanced_chown_but[4].id = group_add_widget (ch_grp, b_group);
 
-    l_mode = label_new (BY + 2, 3, "");
+    l_mode = label_new (BY + 2, 3, NULL);
     group_add_widget (ch_grp, l_mode);
 
     y = BY + 3;
@@ -783,14 +783,15 @@ advanced_chown_dlg_create (void)
         group_add_widget (ch_grp, hline_new (y++, -1, -1));
         advanced_chown_but[i].id = group_add_widget (ch_grp,
                                                      button_new (y,
-                                                                 WIDGET (ch_dlg)->cols / 2 -
+                                                                 WIDGET (ch_dlg)->rect.cols / 2 -
                                                                  advanced_chown_but[i].len,
                                                                  advanced_chown_but[i].ret_cmd,
                                                                  advanced_chown_but[i].flags,
                                                                  advanced_chown_but[i].text, NULL));
         i++;
         advanced_chown_but[i].id = group_add_widget (ch_grp,
-                                                     button_new (y, WIDGET (ch_dlg)->cols / 2 + 1,
+                                                     button_new (y,
+                                                                 WIDGET (ch_dlg)->rect.cols / 2 + 1,
                                                                  advanced_chown_but[i].ret_cmd,
                                                                  advanced_chown_but[i].flags,
                                                                  advanced_chown_but[i].text, NULL));
@@ -801,14 +802,14 @@ advanced_chown_dlg_create (void)
     group_add_widget (ch_grp, hline_new (y++, -1, -1));
     advanced_chown_but[i].id = group_add_widget (ch_grp,
                                                  button_new (y,
-                                                             WIDGET (ch_dlg)->cols / 2 -
+                                                             WIDGET (ch_dlg)->rect.cols / 2 -
                                                              advanced_chown_but[i].len,
                                                              advanced_chown_but[i].ret_cmd,
                                                              advanced_chown_but[i].flags,
                                                              advanced_chown_but[i].text, NULL));
     i++;
     advanced_chown_but[i].id = group_add_widget (ch_grp,
-                                                 button_new (y, WIDGET (ch_dlg)->cols / 2 + 1,
+                                                 button_new (y, WIDGET (ch_dlg)->rect.cols / 2 + 1,
                                                              advanced_chown_but[i].ret_cmd,
                                                              advanced_chown_but[i].flags,
                                                              advanced_chown_but[i].text, NULL));
@@ -830,24 +831,22 @@ advanced_chown_done (gboolean need_update)
 
 /* --------------------------------------------------------------------------------------------- */
 
-static const char *
-next_file (void)
+static const GString *
+next_file (const WPanel *panel)
 {
-    while (!current_panel->dir.list[current_file].f.marked)
+    while (panel->dir.list[current_file].f.marked == 0)
         current_file++;
 
-    return current_panel->dir.list[current_file].fname;
+    return panel->dir.list[current_file].fname;
 }
 
 /* --------------------------------------------------------------------------------------------- */
 
 static gboolean
-try_advanced_chown (const vfs_path_t * p, mode_t m, uid_t u, gid_t g)
+try_advanced_chown (const vfs_path_t *p, mode_t m, uid_t u, gid_t g)
 {
     int chmod_result;
-    const char *fname;
-
-    fname = x_basename (vfs_path_as_str (p));
+    const char *fname = NULL;
 
     while ((chmod_result = mc_chmod (p, m)) == -1 && !ignore_all)
     {
@@ -855,6 +854,8 @@ try_advanced_chown (const vfs_path_t * p, mode_t m, uid_t u, gid_t g)
         int result;
         char *msg;
 
+        if (fname == NULL)
+            fname = x_basename (vfs_path_as_str (p));
         msg = g_strdup_printf (_("Cannot chmod \"%s\"\n%s"), fname, unix_error_string (my_errno));
         result =
             query_dialog (MSG_ERROR, msg, D_ERROR, 4, _("&Ignore"), _("Ignore &all"), _("&Retry"),
@@ -890,6 +891,8 @@ try_advanced_chown (const vfs_path_t * p, mode_t m, uid_t u, gid_t g)
         int result;
         char *msg;
 
+        if (fname == NULL)
+            fname = x_basename (vfs_path_as_str (p));
         msg = g_strdup_printf (_("Cannot chown \"%s\"\n%s"), fname, unix_error_string (my_errno));
         result =
             query_dialog (MSG_ERROR, msg, D_ERROR, 4, _("&Ignore"), _("Ignore &all"), _("&Retry"),
@@ -925,13 +928,13 @@ try_advanced_chown (const vfs_path_t * p, mode_t m, uid_t u, gid_t g)
 /* --------------------------------------------------------------------------------------------- */
 
 static gboolean
-do_advanced_chown (const vfs_path_t * p, mode_t m, uid_t u, gid_t g)
+do_advanced_chown (WPanel *panel, const vfs_path_t *p, mode_t m, uid_t u, gid_t g)
 {
     gboolean ret;
 
     ret = try_advanced_chown (p, m, u, g);
 
-    do_file_mark (current_panel, current_file, 0);
+    do_file_mark (panel, current_file, 0);
 
     return ret;
 }
@@ -939,30 +942,30 @@ do_advanced_chown (const vfs_path_t * p, mode_t m, uid_t u, gid_t g)
  /* --------------------------------------------------------------------------------------------- */
 
 static void
-apply_advanced_chowns (vfs_path_t * vpath, struct stat *sf)
+apply_advanced_chowns (WPanel *panel, vfs_path_t *vpath, struct stat *sf)
 {
     gid_t a_gid = sf->st_gid;
     uid_t a_uid = sf->st_uid;
     gboolean ok;
 
-    if (!do_advanced_chown
-        (vpath, get_mode (), (ch_flags[9] == '+') ? a_uid : (uid_t) (-1),
-         (ch_flags[10] == '+') ? a_gid : (gid_t) (-1)))
+    if (!do_advanced_chown (panel, vpath, get_mode (),
+                            (ch_flags[9] == '+') ? a_uid : (uid_t) (-1),
+                            (ch_flags[10] == '+') ? a_gid : (gid_t) (-1)))
         return;
 
     do
     {
-        const char *fname;
+        const GString *fname;
 
-        fname = next_file ();
-        vpath = vfs_path_from_str (fname);
+        fname = next_file (panel);
+        vpath = vfs_path_from_str (fname->str);
         ok = (mc_stat (vpath, sf) == 0);
 
         if (!ok)
         {
             /* if current file was deleted outside mc -- try next file */
-            /* decrease current_panel->marked */
-            do_file_mark (current_panel, current_file, 0);
+            /* decrease panel->marked */
+            do_file_mark (panel, current_file, 0);
 
             /* try next file */
             ok = TRUE;
@@ -971,14 +974,14 @@ apply_advanced_chowns (vfs_path_t * vpath, struct stat *sf)
         {
             ch_cmode = sf->st_mode;
 
-            ok = do_advanced_chown (vpath, get_mode (),
+            ok = do_advanced_chown (panel, vpath, get_mode (),
                                     (ch_flags[9] == '+') ? a_uid : (uid_t) (-1),
                                     (ch_flags[10] == '+') ? a_gid : (gid_t) (-1));
         }
 
-        vfs_path_free (vpath);
+        vfs_path_free (vpath, TRUE);
     }
-    while (ok && current_panel->marked != 0);
+    while (ok && panel->marked != 0);
 }
 
 /* --------------------------------------------------------------------------------------------- */
@@ -986,7 +989,7 @@ apply_advanced_chowns (vfs_path_t * vpath, struct stat *sf)
 /* --------------------------------------------------------------------------------------------- */
 
 void
-advanced_chown_cmd (void)
+advanced_chown_cmd (WPanel *panel)
 {
     gboolean need_update;
     gboolean end_chown;
@@ -994,7 +997,7 @@ advanced_chown_cmd (void)
     /* Number of files at startup */
     int files_on_begin;
 
-    files_on_begin = MAX (1, current_panel->marked);
+    files_on_begin = MAX (1, panel->marked);
 
     advanced_chown_init ();
 
@@ -1005,7 +1008,7 @@ advanced_chown_cmd (void)
     {                           /* do while any files remaining */
         vfs_path_t *vpath;
         WDialog *ch_dlg;
-        const char *fname;
+        const GString *fname;
         int result;
         int file_idx;
 
@@ -1014,26 +1017,26 @@ advanced_chown_cmd (void)
         need_update = FALSE;
         end_chown = FALSE;
 
-        if (current_panel->marked != 0)
-            fname = next_file ();       /* next marked file */
+        if (panel->marked != 0)
+            fname = next_file (panel);  /* next marked file */
         else
-            fname = selection (current_panel)->fname;   /* single file */
+            fname = panel_current_entry (panel)->fname; /* single file */
 
-        vpath = vfs_path_from_str (fname);
+        vpath = vfs_path_from_str (fname->str);
 
         if (mc_stat (vpath, &sf_stat) != 0)
         {
-            vfs_path_free (vpath);
+            vfs_path_free (vpath, TRUE);
             break;
         }
 
         ch_cmode = sf_stat.st_mode;
 
-        ch_dlg = advanced_chown_dlg_create ();
+        ch_dlg = advanced_chown_dlg_create (panel);
 
-        file_idx = files_on_begin == 1 ? 1 : (files_on_begin - current_panel->marked + 1);
+        file_idx = files_on_begin == 1 ? 1 : (files_on_begin - panel->marked + 1);
         label_set_textv (l_filename, "%s (%d/%d)",
-                         str_fit_to_term (fname, WIDGET (ch_dlg)->cols - 20, J_LEFT_FIT),
+                         str_fit_to_term (fname->str, WIDGET (ch_dlg)->rect.cols - 20, J_LEFT_FIT),
                          file_idx, files_on_begin);
         update_ownership ();
 
@@ -1046,35 +1049,36 @@ advanced_chown_cmd (void)
             break;
 
         case B_ENTER:
-            if (current_panel->marked <= 1)
             {
-                /* single or last file */
-                if (mc_chmod (vpath, get_mode ()) == -1)
-                    message (D_ERROR, MSG_ERROR, _("Cannot chmod \"%s\"\n%s"),
-                             fname, unix_error_string (errno));
-                /* call mc_chown only, if mc_chmod didn't fail */
-                else if (mc_chown
-                         (vpath, (ch_flags[9] == '+') ? sf_stat.st_uid : (uid_t) (-1),
-                          (ch_flags[10] == '+') ? sf_stat.st_gid : (gid_t) (-1)) == -1)
-                    message (D_ERROR, MSG_ERROR, _("Cannot chown \"%s\"\n%s"), fname,
-                             unix_error_string (errno));
+                uid_t uid = ch_flags[9] == '+' ? sf_stat.st_uid : (uid_t) (-1);
+                gid_t gid = ch_flags[10] == '+' ? sf_stat.st_gid : (gid_t) (-1);
 
-                end_chown = TRUE;
-            }
-            else if (!try_advanced_chown
-                     (vpath, get_mode (), (ch_flags[9] == '+') ? sf_stat.st_uid : (uid_t) (-1),
-                      (ch_flags[10] == '+') ? sf_stat.st_gid : (gid_t) (-1)))
-            {
-                /* stop multiple files processing */
-                result = B_CANCEL;
-                end_chown = TRUE;
-            }
+                if (panel->marked <= 1)
+                {
+                    /* single or last file */
+                    if (mc_chmod (vpath, get_mode ()) == -1)
+                        message (D_ERROR, MSG_ERROR, _("Cannot chmod \"%s\"\n%s"),
+                                 fname->str, unix_error_string (errno));
+                    /* call mc_chown only, if mc_chmod didn't fail */
+                    else if (mc_chown (vpath, uid, gid) == -1)
+                        message (D_ERROR, MSG_ERROR, _("Cannot chown \"%s\"\n%s"), fname->str,
+                                 unix_error_string (errno));
 
-            need_update = TRUE;
-            break;
+                    end_chown = TRUE;
+                }
+                else if (!try_advanced_chown (vpath, get_mode (), uid, gid))
+                {
+                    /* stop multiple files processing */
+                    result = B_CANCEL;
+                    end_chown = TRUE;
+                }
+
+                need_update = TRUE;
+                break;
+            }
 
         case B_SETALL:
-            apply_advanced_chowns (vpath, &sf_stat);
+            apply_advanced_chowns (panel, vpath, &sf_stat);
             need_update = TRUE;
             end_chown = TRUE;
             break;
@@ -1084,17 +1088,17 @@ advanced_chown_cmd (void)
             break;
         }
 
-        if (current_panel->marked != 0 && result != B_CANCEL)
+        if (panel->marked != 0 && result != B_CANCEL)
         {
-            do_file_mark (current_panel, current_file, 0);
+            do_file_mark (panel, current_file, 0);
             need_update = TRUE;
         }
 
-        vfs_path_free (vpath);
+        vfs_path_free (vpath, TRUE);
 
-        dlg_destroy (ch_dlg);
+        widget_destroy (WIDGET (ch_dlg));
     }
-    while (current_panel->marked != 0 && !end_chown);
+    while (panel->marked != 0 && !end_chown);
 
     advanced_chown_done (need_update);
 }
