@@ -1,7 +1,7 @@
 /*
    Virtual File System: GNU Tar file system.
 
-   Copyright (C) 2000-2020
+   Copyright (C) 2000-2024
    Free Software Foundation, Inc.
 
    Written by:
@@ -83,7 +83,7 @@ enum
 {
     CPIO_UNKNOWN = 0,           /* Not determined yet */
     CPIO_BIN,                   /* Binary format */
-    CPIO_BINRE,                 /* Binary format, reverse endianity */
+    CPIO_BINRE,                 /* Binary format, reverse endianness */
     CPIO_OLDC,                  /* Old ASCII format */
     CPIO_NEWC,                  /* New ASCII format */
     CPIO_CRC                    /* New ASCII format + CRC */
@@ -139,6 +139,13 @@ typedef struct
     GSList *deferred;           /* List of inodes for which another entries may appear */
 } cpio_super_t;
 
+/*** forward declarations (file scope functions) *************************************************/
+
+static ssize_t cpio_find_head (struct vfs_class *me, struct vfs_s_super *super);
+static ssize_t cpio_read_bin_head (struct vfs_class *me, struct vfs_s_super *super);
+static ssize_t cpio_read_oldc_head (struct vfs_class *me, struct vfs_s_super *super);
+static ssize_t cpio_read_crc_head (struct vfs_class *me, struct vfs_s_super *super);
+
 /*** file scope variables ************************************************************************/
 
 static struct vfs_s_subclass cpio_subclass;
@@ -150,14 +157,6 @@ static off_t cpio_position;
 /*** file scope functions ************************************************************************/
 /* --------------------------------------------------------------------------------------------- */
 
-static ssize_t cpio_find_head (struct vfs_class *me, struct vfs_s_super *super);
-static ssize_t cpio_read_bin_head (struct vfs_class *me, struct vfs_s_super *super);
-static ssize_t cpio_read_oldc_head (struct vfs_class *me, struct vfs_s_super *super);
-static ssize_t cpio_read_crc_head (struct vfs_class *me, struct vfs_s_super *super);
-static ssize_t cpio_read (void *fh, char *buffer, size_t count);
-
-/* --------------------------------------------------------------------------------------------- */
-
 static int
 cpio_defer_find (const void *a, const void *b)
 {
@@ -224,7 +223,7 @@ cpio_free_archive (struct vfs_class *me, struct vfs_s_super *super)
 /* --------------------------------------------------------------------------------------------- */
 
 static int
-cpio_open_cpio_file (struct vfs_class *me, struct vfs_s_super *super, const vfs_path_t * vpath)
+cpio_open_cpio_file (struct vfs_class *me, struct vfs_s_super *super, const vfs_path_t *vpath)
 {
     int fd, type;
     cpio_super_t *arch;
@@ -254,7 +253,7 @@ cpio_open_cpio_file (struct vfs_class *me, struct vfs_s_super *super, const vfs_
         s = g_strconcat (super->name, decompress_extension (type), (char *) NULL);
         tmp_vpath = vfs_path_from_str_flags (s, VPF_NO_CANON);
         fd = mc_open (tmp_vpath, O_RDONLY);
-        vfs_path_free (tmp_vpath);
+        vfs_path_free (tmp_vpath, TRUE);
         if (fd == -1)
         {
             message (D_ERROR, MSG_ERROR, _("Cannot open cpio archive\n%s"), s);
@@ -457,7 +456,7 @@ cpio_create_entry (struct vfs_class *me, struct vfs_s_super *super, struct stat
     if (entry != NULL)
     {
         /* This shouldn't happen! (well, it can happen if there is a record for a
-           file and than a record for a directory it is in; cpio would die with
+           file and then a record for a directory it is in; cpio would die with
            'No such file or directory' is such case) */
 
         if (!S_ISDIR (entry->ino->st.st_mode))
@@ -471,15 +470,8 @@ cpio_create_entry (struct vfs_class *me, struct vfs_s_super *super, struct stat
             entry->ino->st.st_mode = st->st_mode;
             entry->ino->st.st_uid = st->st_uid;
             entry->ino->st.st_gid = st->st_gid;
-#ifdef HAVE_STRUCT_STAT_ST_MTIM
-            entry->ino->st.st_atim = st->st_atim;
-            entry->ino->st.st_mtim = st->st_mtim;
-            entry->ino->st.st_ctim = st->st_ctim;
-#else
-            entry->ino->st.st_atime = st->st_atime;
-            entry->ino->st.st_mtime = st->st_mtime;
-            entry->ino->st.st_ctime = st->st_ctime;
-#endif
+
+            vfs_copy_stat_times (st, &entry->ino->st);
         }
 
         g_free (name);
@@ -603,11 +595,12 @@ cpio_read_bin_head (struct vfs_class *me, struct vfs_s_super *super)
 #ifdef HAVE_STRUCT_STAT_ST_RDEV
     st.st_rdev = u.buf.c_rdev;
 #endif
-    st.st_size = (u.buf.c_filesizes[0] << 16) | u.buf.c_filesizes[1];
-#ifdef HAVE_STRUCT_STAT_ST_MTIM
-    st.st_atim.tv_nsec = st.st_mtim.tv_nsec = st.st_ctim.tv_nsec = 0;
-#endif
-    st.st_atime = st.st_mtime = st.st_ctime = (u.buf.c_mtimes[0] << 16) | u.buf.c_mtimes[1];
+    st.st_size = ((off_t) u.buf.c_filesizes[0] << 16) | u.buf.c_filesizes[1];
+
+    vfs_zero_stat_times (&st);
+
+    st.st_atime = st.st_mtime = st.st_ctime =
+        ((time_t) u.buf.c_mtimes[0] << 16) | u.buf.c_mtimes[1];
 
     return cpio_create_entry (me, super, &st, name);
 }
@@ -676,9 +669,8 @@ cpio_read_oldc_head (struct vfs_class *me, struct vfs_s_super *super)
     u.st.st_rdev = hd.c_rdev;
 #endif
     u.st.st_size = hd.c_filesize;
-#ifdef HAVE_STRUCT_STAT_ST_MTIM
-    u.st.st_atim.tv_nsec = u.st.st_mtim.tv_nsec = u.st.st_ctim.tv_nsec = 0;
-#endif
+
+    vfs_zero_stat_times (&u.st);
     u.st.st_atime = u.st.st_mtime = u.st.st_ctime = hd.c_mtime;
 
     return cpio_create_entry (me, super, &u.st, name);
@@ -757,9 +749,8 @@ cpio_read_crc_head (struct vfs_class *me, struct vfs_s_super *super)
     u.st.st_rdev = makedev (hd.c_rdev, hd.c_rdevmin);
 #endif
     u.st.st_size = hd.c_filesize;
-#ifdef HAVE_STRUCT_STAT_ST_MTIM
-    u.st.st_atim.tv_nsec = u.st.st_mtim.tv_nsec = u.st.st_ctim.tv_nsec = 0;
-#endif
+
+    vfs_zero_stat_times (&u.st);
     u.st.st_atime = u.st.st_mtime = u.st.st_ctime = hd.c_mtime;
 
     return cpio_create_entry (me, super, &u.st, name);
@@ -769,8 +760,8 @@ cpio_read_crc_head (struct vfs_class *me, struct vfs_s_super *super)
 /** Need to CPIO_SEEK_CUR to skip the file at the end of add entry!!!! */
 
 static int
-cpio_open_archive (struct vfs_s_super *super, const vfs_path_t * vpath,
-                   const vfs_path_element_t * vpath_element)
+cpio_open_archive (struct vfs_s_super *super, const vfs_path_t *vpath,
+                   const vfs_path_element_t *vpath_element)
 {
     (void) vpath_element;
 
@@ -810,7 +801,7 @@ cpio_open_archive (struct vfs_s_super *super, const vfs_path_t * vpath,
 /** Remaining functions are exactly same as for tarfs (and were in fact just copied) */
 
 static void *
-cpio_super_check (const vfs_path_t * vpath)
+cpio_super_check (const vfs_path_t *vpath)
 {
     static struct stat sb;
     int stat_result;
@@ -822,8 +813,8 @@ cpio_super_check (const vfs_path_t * vpath)
 /* --------------------------------------------------------------------------------------------- */
 
 static int
-cpio_super_same (const vfs_path_element_t * vpath_element, struct vfs_s_super *parc,
-                 const vfs_path_t * vpath, void *cookie)
+cpio_super_same (const vfs_path_element_t *vpath_element, struct vfs_s_super *parc,
+                 const vfs_path_t *vpath, void *cookie)
 {
     struct stat *archive_stat = cookie; /* stat of main archive */
 
@@ -872,7 +863,7 @@ cpio_read (void *fh, char *buffer, size_t count)
 /* --------------------------------------------------------------------------------------------- */
 
 static int
-cpio_fh_open (struct vfs_class *me, vfs_file_handler_t * fh, int flags, mode_t mode)
+cpio_fh_open (struct vfs_class *me, vfs_file_handler_t *fh, int flags, mode_t mode)
 {
     (void) fh;
     (void) mode;
